//   sha1
// (c) Proteus
// icq: 133575351  email:lawnmower-man@mail.ru 
//
//          Intel-  
//         
// ,       // #
//          (  )

#include<stdio.h>
#include<memory.h>
#include<stdlib.h>
#include<assert.h>

typedef unsigned char byte;
typedef unsigned long dword;
typedef unsigned __int64 dword64;
typedef unsigned int uint;


//! hash: sha
class SHA1{
	union{
		dword abcde[5];		//!<    
		byte output[20];	//!<   
	};
	union{
		byte buffer[64];	//!<    
		dword data[80];		//!<    
	};
	uint size;				//!<  
	uint total_size;		//!<    
	void add(byte);			//!<    

	void init();
	void final(byte[20]);
	void hash_block();		//!<   
public:
	void hash(const char*,byte res[20]);
};


void SHA1::init()
{
	size=total_size=0;
	abcde[0]=0x67452301; 
	abcde[1]=0xefcdab89;
	abcde[2]=0x98badcfe; 
	abcde[3]=0x10325476;
	abcde[4]=0xc3d2e1f0;
}


void SHA1::add(byte x)
{
	total_size++;
	buffer[size++^3]=x; //#
	if(size>=64) hash_block(),size=0;
}

/***
//#     .. 
//#     -  htonl      
//dword SHA1::swap(dword x)
{
	return _rotl((x&0xFF00FF00)>>8 | (x&0x00FF00FF)<<8 ,16);
}
*/

void SHA1::final(byte res[20]) 
{
	uint n;
	dword64 total=dword64(total_size)*8;
	//   
	add(0x80); while(size!=56) add(0); 
	//       
	const byte *t=(byte*)&total;
	for(n=8; n;) add(t[--n]);
	//   ,  
	for(n=0; n<20; n++) res[n]=output[n^3];	//#
}


void SHA1::hash_block() 
{
	uint n;
	dword a=abcde[0];
	dword b=abcde[1];
	dword c=abcde[2];
	dword d=abcde[3];
	dword e=abcde[4];
	//   ()
	for(n=16; n<80; n++) data[n]=_rotl(data[n-3]^data[n-8]^data[n-14]^data[n-16],1);
	//  
	for(n= 0; n<80; n++) {
		dword t,p;
		switch(n/20) {
		case 0: t=(c^d)&b^d;   p=0x5a827999; break;
		case 1:	t=b^c^d;       p=0x6ed9eba1; break;
		case 2: t=b&c|d&(b|c); p=0x8f1bbcdc; break;
		case 3: t=b^c^d;       p=0xca62c1d6; break;
		}
		t+=_rotl(a,5)+e+data[n]+p;
		e=d,d=c,c=_rotl(b,30),b=a,a=t;
	}
	abcde[0]+=a;
	abcde[1]+=b;
	abcde[2]+=c;
	abcde[3]+=d;
	abcde[4]+=e;
}


void SHA1::hash(const char* str,byte res[20])
{
	init();
	while(*str) add(*str++);
	final(res);
}



void main()
{
	assert(sizeof(byte)==1 && sizeof(dword)==4 && sizeof(dword64)==8);

	byte res[20];
	byte test[20]={0xd8,0xf4,0x59,0x03,0x20,0xe1,0x34,0x3a,0x91,0x5b,0x63,0x94,0x17,0x06,0x50,0xa8,0xf3,0x5d,0x69,0x26};
	SHA1().hash("sha",res);
	if(memcmp(res,test,20)) printf("Internal error!\n\a"),getchar();
	printf("done..\n");
}

